import * as React from 'react';
import { Avatar, AvatarProps, avatarSizeValues } from '@fluentui/react-avatar';
import {
  ContactIcon,
  GroupIcon,
  CatIcon,
  IDBadgeIcon,
  CalendarIcon,
  TelemarketerIcon,
  RoomIcon,
  ChatBotIcon,
  SkypeClockIcon,
  SkypeCheckIcon,
  SkypeMinusIcon,
  SkypeArrowIcon,
} from '@fluentui/react-icons-mdl2';
import { StoryExample } from '../utils/StoryExample';
import { Button, SpinButton, Stack, ThemeProvider } from '@fluentui/react';
import { AvatarExamples } from '@fluentui/example-data';
import { useBoolean } from '@fluentui/react-hooks';

const examples = {
  ...AvatarExamples,
  size: avatarSizeValues,
  icon: [
    /* eslint-disable react/jsx-key */
    <GroupIcon />,
    <CatIcon />,
    <CalendarIcon />,
    <RoomIcon />,
    <IDBadgeIcon />,
    <TelemarketerIcon />,
    /* eslint-enable react/jsx-key */
  ],
  badge: [
    'success',
    'warning',
    'error',
    'info',
    { state: 'success', icon: { as: SkypeCheckIcon } },
    { state: 'warning', icon: { as: SkypeClockIcon } },
    { state: 'error', icon: { as: SkypeMinusIcon } },
    { state: 'info', icon: { as: SkypeArrowIcon } },
  ],
} as const;

export const Basic = () => (
  <>
    <StoryExample title="Simple examples">
      <Avatar />
      <Avatar name={examples.name[0]} />
      <Avatar size={40} icon={<IDBadgeIcon />} />
      <Avatar size={72} name={examples.name[0]} image={examples.image[0]} />
    </StoryExample>
    <StoryExample title="Square">
      <Avatar square name="Group" />
      <Avatar square icon={<GroupIcon />} />
    </StoryExample>
    <StoryExample title="Badges">
      <Avatar name={examples.name[1]} badge="warning" />
      <Avatar name={examples.name[2]} badge="success" />
      <Avatar
        name={examples.name[3]}
        image={examples.image[3]}
        badge={{ state: 'success', icon: { as: SkypeCheckIcon } }}
      />
    </StoryExample>
    <StoryExample title="Size">
      <Avatar size={20} name={examples.name[4]} image={examples.image[4]} badge="info" />
      <Avatar size={48} name={examples.name[5]} image={examples.image[5]} badge="success" />
      <Avatar size={96} name={examples.name[6]} image={examples.image[6]} badge="warning" />
    </StoryExample>
    <StoryExample title="Brand color">
      <Avatar colorVariant="brand" name={examples.name[4]} badge="info" />
      <Avatar colorVariant="brand" name={examples.name[5]} icon={examples.icon[5]} badge="success" />
    </StoryExample>
    <StoryExample title="Active/inactive">
      <Stack horizontal wrap tokens={{ childrenGap: 16 }}>
        <Avatar name={examples.name[7]} active="active" />
        <Avatar image={examples.image[8]} active="active" activeDisplay="shadow" />
        <Avatar image={examples.image[9]} active="active" activeDisplay="glow" />
        <Avatar image={examples.image[10]} active="active" activeDisplay="ring-shadow" />
        <Avatar image={examples.image[11]} active="active" activeDisplay="ring-glow" />
        <Avatar image={examples.image[12]} active="inactive" />
      </Stack>
    </StoryExample>
  </>
);

export const AllSizes = () => (
  <>
    <StoryExample title="Image">
      <AvatarExampleList images={examples.image} />
    </StoryExample>
    <StoryExample title="Image, square">
      <AvatarExampleList images={examples.image} square exampleIndex={1} />
    </StoryExample>
    <StoryExample title="Initials">
      <AvatarExampleList names={examples.name} />
    </StoryExample>
    <StoryExample title="Initials, brand color">
      <AvatarExampleList names={examples.name} colorVariant="brand" />
    </StoryExample>
    <StoryExample title="Initials, square">
      <AvatarExampleList names={examples.name} square exampleIndex={1} />
    </StoryExample>
    <StoryExample title="Icon">
      <AvatarExampleList icons={examples.icon} />
    </StoryExample>
    <StoryExample title="Icon, square">
      <AvatarExampleList icons={examples.icon} square exampleIndex={1} />
    </StoryExample>
    <StoryExample title="Icon, brand color">
      <AvatarExampleList icons={examples.icon} colorVariant="brand" exampleIndex={1} />
    </StoryExample>
  </>
);

export const Active = () => (
  <>
    <StoryExample title="ring">
      <AvatarExampleList images={examples.image} active="active" activeDisplay="ring" exampleIndex={2} />
    </StoryExample>
    <StoryExample title="ring-shadow">
      <AvatarExampleList images={examples.image} active="active" activeDisplay="ring-shadow" exampleIndex={3} />
    </StoryExample>
    <StoryExample title="ring-glow">
      <AvatarExampleList images={examples.image} active="active" activeDisplay="ring-glow" exampleIndex={4} />
    </StoryExample>
    <StoryExample title="shadow">
      <AvatarExampleList images={examples.image} active="active" activeDisplay="shadow" exampleIndex={5} />
    </StoryExample>
    <StoryExample title="glow">
      <AvatarExampleList images={examples.image} active="active" activeDisplay="glow" exampleIndex={6} />
    </StoryExample>
    <StoryExample title="inactive">
      <AvatarExampleList images={examples.image} active="inactive" exampleIndex={7} />
    </StoryExample>
  </>
);

export const ActiveAnimation = () => {
  const [active, setActive] = React.useState(false);
  const [size, nextSize, prevSize] = useValueSelectorState(examples.size, 96);
  const [activeDisplay, nextActiveDisplay, prevActiveDisplay] = useValueSelectorState(examples.activeDisplay, 'ring');
  const [display, nextDisplay, prevDisplay] = useValueSelectorState(['image', 'icon', 'label'], 'image');

  React.useEffect(() => {
    const id = setTimeout(() => setActive(true), 500);
    return () => clearTimeout(id);
  }, []);

  return (
    <ThemeProvider>
      <Stack>
        <div style={{ width: 200, height: 200, display: 'flex', alignItems: 'center', justifyContent: 'center' }}>
          <Avatar
            size={size}
            active={active ? 'active' : 'inactive'}
            activeDisplay={activeDisplay}
            name={examples.name[10]}
            image={display === 'image' && examples.image[10]}
            icon={display === 'icon' && <ContactIcon />}
          />
        </div>
        <Stack tokens={{ childrenGap: 8, maxWidth: 220 }}>
          <Button primary onClick={React.useCallback(() => setActive(a => !a), [])}>
            Toggle Active
          </Button>
          <SpinButton
            label="activeDisplay"
            value={activeDisplay}
            onIncrement={nextActiveDisplay}
            onDecrement={prevActiveDisplay}
          />
          <SpinButton label="display" value={display} onIncrement={nextDisplay} onDecrement={prevDisplay} />
          <SpinButton label="size" value={`${size}`} onIncrement={nextSize} onDecrement={prevSize} />
        </Stack>
      </Stack>
    </ThemeProvider>
  );
};

export const CustomSizes = () => (
  <StoryExample title="Custom size">
    <Avatar name={examples.name[11]} badge="success" size={20} tokens={{ width: '13px', height: '13px' }} />
    <Avatar image={examples.image[12]} badge="warning" size={20} tokens={{ width: '21px', height: '21px' }} />
    <Avatar name={examples.name[13]} badge="error" size={32} tokens={{ width: '34px', height: '34px' }} />
    <Avatar image={examples.image[14]} badge="info" size={48} tokens={{ width: '55px', height: '55px' }} />
    <Avatar name={examples.name[15]} badge="warning" size={72} tokens={{ width: '89px', height: '89px' }} />
    <Avatar image={examples.image[16]} badge="success" size={128} tokens={{ width: '144px', height: '144px' }} />
  </StoryExample>
);

export const CustomShape = () => {
  return (
    <>
      <StoryExample title="Custom shape">
        <AvatarExampleList
          icon={<ChatBotIcon />}
          tokens={{
            width: 'calc(var(--avatar-height) * 1.125)',
            background: `url('${examples.hexagon}') 0px/contain no-repeat`,
            borderRadius: '0',
          }}
        />
      </StoryExample>
    </>
  );
};

export const AvatarPlayground = () => {
  const [nameAndImage, nextNameAndImage, prevNameAndImage] = useValueSelectorState(examples.nameAndImage);

  const propSelectors = [
    useValueSelector('size', useValueSelectorState(examples.size, 96), true),
    useValueSelector('square', useValueSelectorState([true, false])),
    useValueSelector('badge', useValueSelectorState(examples.badge), false, badgeToString),
    useValueSelector('name', [nameAndImage.name, nextNameAndImage, prevNameAndImage], true),
    useValueSelector('image', [nameAndImage.image, nextNameAndImage, prevNameAndImage], true, getFilenameFromUrl),
    useValueSelector('icon', useValueSelectorState(examples.icon), false, iconToString),
    useValueSelector('colorVariant', useValueSelectorState(examples.colorVariant)),
    useValueSelector('active', useValueSelectorState(['active', 'inactive'] as const)),
    useValueSelector('activeDisplay', useValueSelectorState(examples.activeDisplay)),
  ];

  // Build an AvatarProps object with the selected property values
  const propValues: AvatarProps = {};
  propSelectors.forEach(({ assignValue }) => assignValue(propValues));

  return (
    <div style={{ display: 'flex' }}>
      <div style={{ width: '225px', height: '225px', display: 'flex', alignItems: 'center', justifyContent: 'center' }}>
        <Avatar {...propValues} />
      </div>
      <div style={{ display: 'flex', flexDirection: 'column', fontFamily: 'Consolas, monospaced', fontSize: '14px' }}>
        {...propSelectors.map(p => p.renderSelector())}
        <div style={{ display: 'flex', flexDirection: 'column', marginTop: '20px' }}>
          {`<Avatar `}
          {...propSelectors.map(p => p.renderValue())}
          {`/>`}
        </div>
      </div>
    </div>
  );
};

//
// Helpers
//

/**
 * Generate a list of Avatars with sample properties
 */
const AvatarExampleList: React.FC<AvatarProps & {
  names?: readonly string[];
  images?: readonly string[];
  icons?: readonly JSX.Element[];
  exampleIndex?: number;
}> = props => {
  const { names, images, icons, exampleIndex = 0, ...restOfProps } = props;
  const offset = exampleIndex * examples.size.length;

  return (
    <Stack wrap horizontal tokens={{ childrenGap: 24 }}>
      {examples.size.map((size, i) => (
        <Avatar
          key={size}
          size={size}
          name={names && names[(i + offset) % names.length]}
          image={images && images[(i + offset) % images.length]}
          icon={icons && icons[(i + offset) % icons.length]}
          badge={examples.badge[(i + offset) % examples.badge.length]}
          {...restOfProps}
        />
      ))}
    </Stack>
  );
};

const iconToString = (icon: JSX.Element | undefined): string => `<${icon?.type.displayName} />`;
const badgeToString = (badge: typeof examples.badge[number] | undefined): string =>
  typeof badge === 'object' ? `{ state: '${badge.state}', icon: { as: ${badge.icon.as.displayName} } }` : `${badge}`;
const getFilenameFromUrl = (url: string) => url.substring(url.lastIndexOf('/') + 1);

type ValueSelectorState<T> = [/*value:*/ T, /*next:*/ () => void, /*prev:*/ () => void];

/**
 * Select a value from an array of values, with next/previous methods
 */
const useValueSelectorState = function<T>(values: readonly T[], initialValue: T = values[0]): ValueSelectorState<T> {
  const count = values.length;
  const [index, setIndex] = React.useState<number>(() => values.indexOf(initialValue));
  const next = React.useCallback(() => setIndex(i => (i + 1) % count), [count]);
  const prev = React.useCallback(() => setIndex(i => (i - 1 + count) % count), [count]);
  return [values[index], next, prev];
};

/**
 * Create a selector UI for a property value, allowing the user to toggle among the available property values
 */
const useValueSelector = <Prop extends keyof AvatarProps>(
  name: Prop,
  [value, next, prev]: ValueSelectorState<AvatarProps[Prop]>,
  initialEnabled: boolean = false,
  valueToString: (v: AvatarProps[Prop] | undefined) => string = v => `${v}`,
) => {
  const [enabled, { toggle: toggleEnabled }] = useBoolean(initialEnabled);

  return {
    /** Assign this property's value to the given props object, if the property is set */
    assignValue: (props: AvatarProps) => enabled && (props[name] = value),

    /** Render the UI to select the property value */
    renderSelector: () => (
      <div style={{ opacity: !enabled ? '50%' : undefined }}>
        <button onClick={enabled ? prev : toggleEnabled}>&lt;</button>
        <button onClick={enabled ? next : toggleEnabled}>&gt;</button>
        <input id={`prop_${name}`} type="checkbox" onChange={toggleEnabled} checked={enabled} />
        <label htmlFor={`prop_${name}`}>{`${name}: ${enabled ? `${valueToString(value)}` : `(unset)`}`}</label>
      </div>
    ),

    /** Render a span with propName="propValue" inside, if the property is set */
    renderValue: () => {
      const quotes = typeof value === 'string' ? '""' : '{}';
      return enabled && <span>{`${name}=${quotes[0]}${valueToString(value)}${quotes[1]}`}</span>;
    },
  };
};
